

/*
 *	This daemon process is used to maintain a user space socket state.
 *
 *	We are going to make this work in the kernel. But now we employ the 
 *	easiest way
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/epoll.h>
#include <syslog.h>
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#include <vector>
#include <list>

#include "fastopen_daemon.h"

using namespace std;

/***************************************************************
 *                   Err informations                          *
 *-------------------------------------------------------------*
 *      This section records the error string information      *
 *      because we can not use standard input/output to        *
 *      tell user directly                                     *
 ***************************************************************/

extern "C" {
static void open_err_log(const char* prg_name)
{
	openlog(prg_name, LOG_NDELAY, LOG_USER);
}


extern "C" {
#define fastopen_err(fmt, ...)\
	fprintf(stderr, fmt, ##__VA_ARGS__)
#define fastopen_debug(fmt, ...) \
	fprintf(stderr, fmt, ##__VA_ARGS__)
}


static void close_log()
{
	closelog();
}
}

/***************************************************************
 *                   Service Path and sock table               *
 ***************************************************************/

struct PidEntry;
struct SockEntry;



/*
 *	Record the fastopen socket state
 */
struct SockEntry {
	int state;
	int refcnt;
	int gid;
	list <PidEntry*> ref_procs;
	/*
	 *	There might be processes waiting for the state change,
	 *	and chain them on the socket entry
	 */
	list <struct sockaddr_un> wait_list;
	SockEntry(int s, int rf, int g) :
		state(s), refcnt(rf), gid(g) {
		}
};

/*
 *	Record the registered process's information
 */
struct PidEntry {
	pid_t pid;
	list <SockEntry*> used_socks;
	PidEntry(pid_t p): pid(p) {}
};

static vector<SockEntry> sock_table;
static list<SockEntry*> free_list;
static list<PidEntry> pid_list;

static void init_sock_table()
{
	sock_table.reserve(MAX_FASTOPEN_SOCK);
	for (int i = 0; i < MAX_FASTOPEN_SOCK; i++) {
		sock_table.push_back(SockEntry(FASTOPEN_SOCK_INVALID, 0, i));
		free_list.push_back(&sock_table[i]);
	}
}

static SockEntry* get_sock()
{
	if (free_list.empty()) {
		return NULL;
	}
	SockEntry* ent = free_list.front();
	free_list.pop_front();
	ent->state = FASTOPEN_SOCK_IDLE;
	return ent;
}

static void put_sock(SockEntry* ent)
{
	free_list.push_back(ent);
}

static SockEntry* find_sock(int gsid)
{
	if (gsid < 0) return NULL;
	return &sock_table[gsid];
}

static int get_sock_state(int gsid)
{
	SockEntry* ent = find_sock(gsid);
	if (ent == NULL) return -1;
	return ent->state;
}

static int wait_for_connect(int gsid, struct sockaddr_un* addr)
{
	SockEntry* ent = find_sock(gsid);
	if (ent == NULL) return -1;
	if (ent->state == FASTOPEN_SOCK_CONNECTED) {
		return 0;
	} else {
		printf("enter wait list\n");
		ent->wait_list.push_back(*addr);
		return -1;
	}
}

static int set_sock_state(int gsid, int new_state)
{
	SockEntry* ent = find_sock(gsid);
	if (ent == NULL) return -1;
	int old_state = ent->state;
	ent->state = new_state;
	return old_state;
}

static int cmp_set_sock_state(int gsid, int cmp_state, int new_state)
{
	SockEntry* ent = find_sock(gsid);
	if (ent == NULL) return -1;
	int old_state = ent->state;
	if (old_state == cmp_state)
		ent->state = new_state;
	return old_state;
}

static void invoke_wait_sock(int notify_fd, int gsid, 
		void (*invoke)(int, int, struct sockaddr_un&))
{
	SockEntry* ent = find_sock(gsid);
	if (ent == NULL) return ;
	printf("entry of the fastopen state %p %d\n", ent, ent->gid);
	if (ent->state == FASTOPEN_SOCK_CONNECTED || ent->state == FASTOPEN_SOCK_INVALID) {
		/*
		 *	Invoke the wait list 
		 */
		list <struct sockaddr_un>& wait_list = ent->wait_list;
		printf("do we come to this:%d\n", wait_list.size());
		while (!wait_list.empty()) {
			struct sockaddr_un& wait_addr = wait_list.front();
			invoke(notify_fd, ent->state, wait_addr);
			wait_list.pop_front();
		}
	}
}

static PidEntry* find_proc_ent(pid_t p)
{
	for (list<PidEntry>::iterator it = pid_list.begin();
			it != pid_list.end(); ++it) {
		if (it->pid == p)
			return &(*it);
	}
	return NULL;
}


static int bind_sock(PidEntry* pent, SockEntry* ent)
{
	//
	//	FIXME:
	//		We don't handle rebind issue currently. We
	//		assume it will not happen now
	//
	++ent->refcnt;
	ent->ref_procs.push_back(pent);
	pent->used_socks.push_back(ent);
	return ent->gid;
}

static	int bind_sock(pid_t pid, int gsid)
{
	SockEntry* ent = gsid == FASTOPEN_NEW_SOCK? get_sock(): find_sock(gsid);
	PidEntry* pent = find_proc_ent(pid);
	if (pent == NULL && gsid == FASTOPEN_NEW_SOCK) {
		put_sock(ent);
		return -1;
	}
	return bind_sock(find_proc_ent(pid), ent);
}

static void unbind_sock(PidEntry* pent, SockEntry* ent)
{
	ent->ref_procs.remove(pent);
	if (--ent->refcnt == 0)
		put_sock(ent);
}

static void unbind_sock(pid_t pid, int gsid)
{
	SockEntry* ent = find_sock(gsid);
	if (ent == NULL) return ;
	PidEntry* pent = find_proc_ent(pid);
	if (pent == NULL) return;
	pent->used_socks.remove(ent);
	unbind_sock(pent, ent);
}

static void unbind_all_sock(PidEntry* pent)
{
	while (!pent->used_socks.empty()){
		unbind_sock(pent, pent->used_socks.front());
		pent->used_socks.pop_front();
	}
}

static void unbind_all_sock(pid_t pid)
{
	PidEntry* pent = find_proc_ent(pid);
	if (pent == NULL) return ;
	unbind_all_sock(pent);
}

static void del_proc(pid_t pid)
{
	for (list <PidEntry>::iterator it = pid_list.begin();
			it != pid_list.end(); ++it)
		if (it->pid == pid) {
			unbind_all_sock(&(*it));
			pid_list.erase(it);
			return;
		}
}


static void add_proc(pid_t pid)
{
	pid_list.push_back(PidEntry(pid));
}

/*
 *	Duplicate proc1's used sock to proc2
 */
static void dup_proc(pid_t p1, pid_t p2)
{
	PidEntry* pent1 = find_proc_ent(p1);
	PidEntry* pent2 = find_proc_ent(p2);

	if (pent1 == NULL || pent2 == NULL)
		return ;
	for (list <SockEntry*>::iterator it = pent1->used_socks.begin();
			it != pent1->used_socks.end(); ++it)
		bind_sock(pent2, *it);
}

/***************************************************************
 *                Exported Interfaces                          *
 *-------------------------------------------------------------*
 *      This daemon employs the unix domain socket to chat     *
 *      with user. It tells what the user wrapper library      *
 *      current state of a socket			       *
 ***************************************************************/

static void do_invoke_sock(int out_fd, int cur_state, struct sockaddr_un& wait_addr)
{
	struct FASTOPENRet ret;
	ret.is_valid = 1;
	ret.state = cur_state;
	printf("invoke wait socks\n");
	printf("addresss: %s\n", wait_addr.sun_path);
	sendto(out_fd, &ret, sizeof(ret), 0, (struct sockaddr*) &wait_addr, sizeof(wait_addr));
}

static void fastopen_invoke_sock(int out_fd, struct FASTOPENCmd* cmd)
{
	invoke_wait_sock(out_fd, cmd->gsockfd, do_invoke_sock);
}

static void fastopen_add_proc(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	add_proc(cmd->pid1);
	ret->should_ret = 0;
}

static void fastopen_del_proc(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	del_proc(cmd->pid1);
	ret->should_ret = 0;
}

static void fastopen_dup_proc(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	dup_proc(cmd->pid2, cmd->pid1);
	ret->should_ret = 0;
}

static void fastopen_add_sock(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	ret->should_ret = 1;
	ret->is_valid = 1;
	ret->state = bind_sock(cmd->pid1, cmd->gsockfd);
	printf("state = %d\n", ret->state);
}

static void fastopen_del_sock(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	unbind_sock(cmd->pid1, cmd->gsockfd);
	ret->should_ret = 0;
}

static void fastopen_get_sock_state(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	ret->should_ret = 1;
	ret->is_valid = 1;
	ret->state = get_sock_state(cmd->gsockfd);
}

static void fastopen_wait_sock_state(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret, struct sockaddr_un* addr)
{
	int state = get_sock_state(cmd->gsockfd);
	if (state == FASTOPEN_SOCK_INVALID || state == FASTOPEN_SOCK_CONNECTED) {
		ret->should_ret = 1;
		ret->is_valid = 1;
		ret->state = state;
	} else {
		ret->should_ret = 0;
		wait_for_connect(cmd->gsockfd, addr);
	}
}


static void fastopen_set_sock_state(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	set_sock_state(cmd->gsockfd, cmd->state);
	ret->is_valid = 1;
	ret->should_ret = 0;
}

static void fastopen_cmp_set_sock_state(struct FASTOPENCmd* cmd, struct FASTOPENRet* ret)
{
	ret->is_valid = 1;
	ret->should_ret = 1;
	ret->state = cmp_set_sock_state(cmd->gsockfd, cmd->cmp_state, cmd->state);
}


/***************************************************************
 *			Inner implementation                   *
 ***************************************************************/

/*
 *	Debug usage:
 */
static const char* fastopen_str_state(int type)
{
	switch(type) {
		case FASTOPEN_SOCK_INVALID:
			return "invalid";
		case FASTOPEN_SOCK_IDLE:
			return "idle";
		case FASTOPEN_SOCK_CONNECTING:
			return "connecting";
		case FASTOPEN_SOCK_CONNECTED:
			return "connected";
		default:
			return "no such state";
	}
}

static void dump_single_sock(SockEntry* ent)
{
	fastopen_debug("\tgid:%d\n\trefcnt: %d\n\tstate:%s\n",
			ent->gid, ent->refcnt, fastopen_str_state(ent->state));
	for (list <PidEntry*>::iterator it = ent->ref_procs.begin();
			it != ent->ref_procs.end(); ++it)
		fastopen_debug("\tpid-%u\n", (*it)->pid);
}

static void dump_sock_table()
{
	fastopen_debug("sock free count: %d\n", free_list.size());
	fastopen_debug("sock entries: %d\n", sock_table.size() - free_list.size());
	for (uint32_t i = 0; i < sock_table.size(); i++) {
		if (sock_table[i].refcnt > 0) {
			dump_single_sock(&sock_table[i]);
		}
	}
}

static void dump_single_proc_entry(PidEntry& pent)
{
	fastopen_debug("\tpid: %u\n", pent.pid);
	for (list <SockEntry*>::iterator it = pent.used_socks.begin();
			it != pent.used_socks.end(); ++it)
		fastopen_debug("\t sock id:%d\n", (*it)->gid);
}

static void dump_proc_table()
{
	fastopen_debug("proc count: %d\n", pid_list.size());
	for (list <PidEntry>::iterator it = pid_list.begin();
			it != pid_list.end(); ++it)
		dump_single_proc_entry(*it);
}

static int do_fastopen_cmd(int fd)
{
	struct FASTOPENCmd cmd;
	struct FASTOPENRet ret;
	struct sockaddr_un addr;
	socklen_t addrlen = sizeof(struct sockaddr_un);
	int size = recvfrom(fd, &cmd, sizeof(cmd), 0, (struct sockaddr*) &addr, &addrlen);

	if (size < 0) {
		//	error occurs
		//	This automatically removes the socket from the
		//	epoll list
		fastopen_err("error occurs during processing fastopen command: %s\nabort\n", strerror(errno));
		return -errno;
	}

	fastopen_debug("receive a command of %d, sock %d, state %d, pid %u\n",
			cmd.op, cmd.gsockfd, cmd.state, cmd.pid1);

	switch(cmd.op) {
		case FASTOPEN_ADD_PROC:
			fastopen_add_proc(&cmd, &ret);
			break;
		case FASTOPEN_DEL_PROC:
			fastopen_del_proc(&cmd, &ret);
			break;
		case FASTOPEN_DUP_PROC:
			fastopen_dup_proc(&cmd, &ret);
			break;
		case FASTOPEN_ADD_SOCK:
			fastopen_add_sock(&cmd, &ret);
			break;
		case FASTOPEN_DEL_SOCK:
			fastopen_del_sock(&cmd, &ret);
			break;
		case FASTOPEN_GET_SOCK_STATE:
			fastopen_get_sock_state(&cmd, &ret);
			break;
		case FASTOPEN_WAIT_SOCK_STATE:
			//
			//	Currently this is only for waiting for connected or
			//	closed
			//
			fastopen_wait_sock_state(&cmd, &ret, (struct sockaddr_un*) &addr);
			break;
		case FASTOPEN_SET_SOCK_STATE:
			fastopen_set_sock_state(&cmd, &ret);
			fastopen_invoke_sock(fd, &cmd);
			break;
		case FASTOPEN_CMP_SET_SOCK_STATE:
			fastopen_cmp_set_sock_state(&cmd, &ret);
			fastopen_invoke_sock(fd, &cmd);
			break;
		case FASTOPEN_DEBUG_PROC_LIST:
			ret.should_ret = 0;
			dump_proc_table();
			break;
		case FASTOPEN_DEBUG_SOCK_LIST:
			ret.should_ret = 0;
			dump_sock_table();
			break;
	}
	if (ret.should_ret) {
		int r = sendto(fd, &ret, sizeof(ret), 0, (struct sockaddr*) &addr, addrlen);
		printf ("send to %s %d %s\n", addr.sun_path, r, strerror(errno));
	}
	return 1;
}



static int do_fastopen_poll(int serv)
{
	int ret = 0;
	while (1) {
		ret = do_fastopen_cmd(serv);
		if (ret <= 0) {
			break;
		}
	}
	return ret;
}


static int fastopen_loop(const char* serv)
{
	int serv_fd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (serv_fd < 0) {
		fastopen_err("can not open fastopen daemon sock: %s\n", strerror(errno));
		return -1;
	}

	if (remove(kFASTOPEN_SERV_PATH) != 0 && errno != ENOENT) {
		fastopen_err("error in recreating daemon sock file: %s\n", strerror(errno));
		return -1;
	}

	struct sockaddr_un serv_addr;
	serv_addr.sun_family = AF_UNIX;
	strncpy(serv_addr.sun_path, kFASTOPEN_SERV_PATH, sizeof(serv_addr.sun_path));

	if (bind(serv_fd, (struct sockaddr*) &serv_addr, sizeof(serv_addr)) < 0) {
		fastopen_err("can not bind fastopen daemon sock: %s\n", strerror(errno));
		return -1;
	}
	if (chmod(serv_addr.sun_path, 0777) < 0) {
		fastopen_err("can not set permission for fastopen daemon sock: %s\n", strerror(errno));
		return -1;
	}
	return do_fastopen_poll(serv_fd);
}



int main(int argc, char** argv)
{
	open_err_log(argv[0]);
	atexit(close_log);

	if (0 && daemon(1, 1) < 0) {
		fastopen_err("can not setup daemon process: %s\n", strerror(errno));
		abort();
	}
	fastopen_debug("entering the sock status\n");
	init_sock_table();
	return fastopen_loop(argv[0]);
}



