
#ifdef _DEBUG
#include <assert.h>


#define DEBUG(fmt, ...) fprintf(stderr, "DEBUG: %s-%d: "fmt, __FILE__, __LINE__, ##__VA_ARGS__)
#define ASSERT(x) assert(x)
#else
#define DEBUG(x, ...) 0
#define ASSERT(x) 0
#endif

#include "fastopen_kernel_def.h"
#include "fastopen_daemon.h"

#include <pthread.h>
#include <signal.h>
#include <netinet/ip.h>
#include <dlfcn.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/un.h>
#include <poll.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

/*
FILE* fout;
static void logs()
{
	fout  = fopen("log", "w");
	if (!fout) {
		fprintf(stderr, "fastopen_wrapper: can not open log\n");
	}
}
*/
/*
 *  NOTE:
 *    Should we please change the in kernel implementation
 *    option to the userspace hash table implementation?
 *
 *    pros:
 *      easy to maintain sender's status.
 *
 *    cons:
 *      cause trouble to receiver side, need communication
 *      between module and user-space, might be slow and
 *      not easy
 *
 *
 *    design rationale to solve fork problem:
 *
 *	1. client side
 *		we use the daemon process to keep a global socket state table such
 *		that even we call fork we still have consistent state
 *		the timers and locks must be reinitialized since the fork would bring
 *		strange things
 *
 *	2. server side
 *		in server the sock table is not very important, we just ensure that we
 *		don't call recv on socket by two different threads, that would not ensure
 *		the consistency. Besides, the special data received by SYN packet will be
 *		stored in kernel, so only any processes try to fetch the data could 
 *		correctly determine whether they should do this kind of special fetch any
 *		more, so no need to communicate with the daemon process
 */

/***************************************************************
 *                Instrument syscall Management                *
 *-------------------------------------------------------------*
 *      instrument the libc syscall wrapper to implement the   *
 *      syscall redirection                                    *
 ***************************************************************/

/*
 *  The standard call written in C will use the libc wrapper,
 *  we just instead replace this to allow upper level system
 *  call redirection
 */
static void *libc_handle = NULL;

/*
 *  System calls for the sockets
 *
 *    During an FASTOPEN connection setup, we don't connect until
 *    there is data going out or it has been a long time after
 *    the client explicitly call the connect method
 *
 *    We want to intercept the first data packet sending from
 *    the user side. The packet data will be append to the 
 *    first SYN frame. If the client calls the connect then
 *    don't call any send method, then there will not be any
 *    data from the client going out, thus we have to setup
 *    the timeout mechanism to handle this
 *
 *    These amount of data specially append to the SYN would
 *    be specially received at the server side, such that we
 *    also need to intercept the server side recv option: if
 *    there is such kind of client data pending, then we just
 *    read them firstly
 *
 */

#if 0
static int (*real_socket) (int, int, int);
static int (*real_connect) (int, const struct sockaddr *, socklen_t);
static ssize_t(*real_send) (int, const void *, size_t, int);
//static ssize_t (*real_sendto) (int, const void*, size_t, int, const struct sockaddr*, socklen_t);
static ssize_t(*real_sendmsg) (int, const struct msghdr *, int);
static ssize_t(*real_write) (int, const void *, size_t);
static ssize_t(*real_writev) (int, const struct iovec *, int);
static ssize_t(*real_readv) (int, const struct iovec *, int);
static ssize_t(*real_recv) (int, void *, size_t, int);
//static ssize_t (*real_recvfrom) (int, void*, size_t, int, const struct sockaddr*, socklen_t);
static ssize_t(*real_recvmsg) (int, struct msghdr *, int);
static ssize_t(*real_read) (int, void *, size_t);
static int (*real_accept) (int, struct sockaddr *, socklen_t *);
static int (*real_accept4) (int, struct sockaddr *, socklen_t *, int);
static int (*real_getsockopt) (int, int, int, void *, socklen_t *);
static int (*real_setsockopt) (int, int, int, const void *, socklen_t);
static int (*real_close) (int);
static int (*real_getpeername) (int sockfd, struct sockaddr * addr,
				socklen_t * addrlen);
static int (*real_open) (const char*, int, mode_t);
static int (*real_creat) (const char*, mode_t);
//
//  @TODO:
//    Add wrapper for the asynchronous IO operations:
//      e.g. the application might do:
//        io_submit to query some page, while doing some computation during this
//        process. If this kind of io_submit occurs at the beginning, we need to
//        capture it and append the data to the SYN packet
//

//static int (*real_recv) (int, char*, int, int);
//static int (*real_read) (int, char*, int);
//static int (*real_setsockopt) (int, int, int, void*, socklen_t);
//static int (*real_getsockopt) (int, int, int, void*, socklen_t*);
//static int (*real_bind) (int, const struct sockaddr*, socklen_t);
//static int (*real_listen) (int, int);
//static int (*real_accept) (int, struct sockaddr*, socklen_t);

static void libc_init()
{
	libc_handle = dlopen("/lib/libc.so.6", RTLD_NOW);
	real_socket = dlsym(libc_handle, "socket");
	real_connect = dlsym(libc_handle, "connect");
	real_send = dlsym(libc_handle, "send");
	real_write = dlsym(libc_handle, "write");
	real_sendmsg = dlsym(libc_handle, "sendmsg");
	real_write = dlsym(libc_handle, "write");
	real_writev = dlsym(libc_handle, "writev");
	real_readv = dlsym(libc_handle, "readv");
	real_recv = dlsym(libc_handle, "recv");
	// real_recvfrom = dlsym(libc_handle, "recvfrom");
	real_recvmsg = dlsym(libc_handle, "recvmsg");
	real_read = dlsym(libc_handle, "read");
	real_getsockopt = dlsym(libc_handle, "getsockopt");
	real_setsockopt = dlsym(libc_handle, "setsockopt");
	real_accept = dlsym(libc_handle, "accept");
	real_accept4 = dlsym(libc_handle, "accept4");
	real_close = dlsym(libc_handle, "close");
	real_getpeername = dlsym(libc_handle, "getpeername");
	real_open = dlsym(libc_handle, "open");
	real_creat = dlsym(libc_handle, "creat");
}
#endif






/***************************************************************
 *                      FASTOPEN workers                           *
 *-------------------------------------------------------------*
 *  The FASTOPEN calls that do the real work. They are called in   *
 *  syscall wrappers.                                          *
 ***************************************************************/
/*
 *  A socket table structure
 */
#define MAX_SOCK_FD 65536

/*
 *  FASTOPEN socket status (user space attribute is enough, it's all right
 *    to make things in kernel, but that should work after we prove it
 *    all right in application experiments)
 *
 *  @state  socket's state, invalid (not used), idle, connecting, connected
 *          for portablity, we should use atomic_t to wrap int. But we just
 *          want this running on X86 machines, so int is sufficient
 *  @type   an FASTOPEN client socket or server socket. It's distinguished by
 *          calling "connect" or "accept". Default to NONE, since it's not
 *          determined on creating.
 */
#define FASTOPEN_SOCK_INVALID         0
#define FASTOPEN_SOCK_IDLE            1
#define FASTOPEN_SOCK_CONNECTING      2
#define FASTOPEN_SOCK_CONNECTED       3

#define FASTOPEN_NONE            0
#define FASTOPEN_SOCK	     (1 << 0)
#define FASTOPEN_CLIENT          (1 << 1)
#define FASTOPEN_SERVER          (1 << 2)

/*
 *	@NOTE:
 *		we don't use pthread mutex to protect the state
 *		there is no need.
 *
 *		the most useful field is 'no_special'. This will
 *		recover send/recv to normal behavior
 */
static struct fastopen_sock_status {
	int type;
	int global_id;
	int no_special;
} fastopen_sock_table[MAX_SOCK_FD];

int fastopen_to_global(int fd)
{
	return fastopen_sock_table[fd].global_id;
}

/*
 *	FASTOPEN daemon communication protocols:
 *
 *	serveral commands:
 *
 *	FASTOPEN_ADD_PROC:
 *		this is called on create of the process, except for the first process,
 *		it should be sent in pthread_atfork's child entry
 *	FASTOPEN_DEL_PROC:
 *		this is called on end of the process, it should be called in atexit,
 *		the original process must register it
 *	FASTOPEN_DUP_PROC:
 *		this is called on fork of the process, it should be called in
 *		pthread_atfork's child handle
 *	FASTOPEN_ADD_SOCK:
 *		this is called on creation of the socket, it should be called in socket
 *		wrapper
 *	FASTOPEN_DEL_SOCK:
 *		this is called on close of the socket fd. We must judge whether it's a
 *		socket or not
 *	FASTOPEN_WAIT_SOCK_STATE:
 *		this is called in the fastopen client's sender, it must wait for the connect
 *		process to finish then it can proceed to send data
 *	FASTOPEN_GET_SOCK_STATE:
 *		this is called in the fastopen clients' sender. it can acquire the current
 *		state of the socket to determine whether to connect
 *	FASTOPEN_SET_SOCK_STATE:
 *		this is called either in the fastopen client's sender or the slow connection
 *		thread, it is used to set the socket state and invoke any threads pending
 *		on waiting for this change.
 *	FASTOPEN_CMP_SET_SOCK_STATE:
 *		this is called either in the fastopen client's sender or the slow connection
 *		thread, it is used to compete the connection role. Those who can connect
 *		will get 'FASTOPEN_SOCK_IDLE' on return. The others will get 
 *		'FASTOPEN_SOCK_CONNNECTING', and senders should wait on this and the slow
 *		connection thread should proceed to next socket
 *
 */

static int fastopen_cmd(struct FASTOPENCmd *cmd, struct FASTOPENRet *ret)
{
	int retval = 0;
	int fastopen_daemon_fd = real_socket(AF_UNIX, SOCK_DGRAM, 0);
	if (fastopen_daemon_fd < 0) {
		fprintf(stderr, "can not connect to fastopen daemon: %s\n",
			strerror(errno));
		exit(1);
	}

	struct sockaddr_un fastopen_daemon_local_addr;
	memset(&fastopen_daemon_local_addr, 0, sizeof(fastopen_daemon_local_addr));
	fastopen_daemon_local_addr.sun_family = AF_UNIX;
	sprintf(fastopen_daemon_local_addr.sun_path, "/tmp/fastopen-client.%u.%lu",
		getpid(), pthread_self());
	if (unlink(fastopen_daemon_local_addr.sun_path) != 0 && errno != ENOENT) {
		fprintf(stderr, "can not clean previous local bind file: %s\n",
			strerror(errno));
		exit(2);
	}
	if (bind(fastopen_daemon_fd, (struct sockaddr *)&fastopen_daemon_local_addr,
		 sizeof(struct sockaddr_un)) < 0) {
		fprintf(stderr, "can not bind to local address: %s\n",
			strerror(errno));
		exit(1);
	}
	if (chmod(fastopen_daemon_local_addr.sun_path, 0777) < 0) {
		fprintf(stderr, "can not change permission: %s\n",
			strerror(errno));
		exit(1);
	}
	struct sockaddr_un daemon_addr;
	daemon_addr.sun_family = AF_UNIX;
	strcpy(daemon_addr.sun_path, kFASTOPEN_SERV_PATH);
	if (real_connect
	    (fastopen_daemon_fd, (struct sockaddr *)&daemon_addr,
	     sizeof(daemon_addr)) < 0) {
		fprintf(stderr, "can not connect to fastopen daemon: %s\n",
			strerror(errno));
		exit(1);
	}

	if (real_send(fastopen_daemon_fd, cmd, sizeof(*cmd), 0) != sizeof(*cmd)) {
		retval = -1;
		goto out;
	}
	if (ret == NULL) {
		goto out;
	} else if (real_recv(fastopen_daemon_fd, ret, sizeof(*ret), 0) !=
		   sizeof(*ret)) {
		retval = -1;
		goto out;
	}
 out:
	//DEBUG("closing the daemon: %d\n", retval);
	real_close(fastopen_daemon_fd);
	unlink(fastopen_daemon_local_addr.sun_path);
	return retval;
}

static void fastopen_add_proc()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_ADD_PROC,
		.pid1 = getpid()
	};
	fastopen_cmd(&cmd, NULL);
}

static void fastopen_del_proc()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DEL_PROC,
		.pid1 = getpid()
	};
	fastopen_cmd(&cmd, NULL);
}

static int fastopen_dup_proc()
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DUP_PROC,
		.pid1 = getpid(),
		.pid2 = getppid()
	};
	return fastopen_cmd(&cmd, NULL);
}

static int fastopen_add_sock(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_ADD_SOCK,
		.gsockfd = fastopen_to_global(fd),
		.pid1 = getpid()
	};
	struct FASTOPENRet ret;
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid))
		return -1;
	fastopen_sock_table[fd].global_id = ret.state;
	return ret.state;
}

static int fastopen_del_sock(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_DEL_SOCK,
		.gsockfd = fastopen_to_global(fd),
		.pid1 = getpid()
	};
	return fastopen_cmd(&cmd, NULL);
}

static int fastopen_get_sock_state(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_GET_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
	};
	struct FASTOPENRet ret;
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid))
		return -1;
	return ret.state;
}

static int fastopen_wait_connect(int fd)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_WAIT_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
	};
	struct FASTOPENRet ret;
	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid))
		return -1;
	return ret.state;
}

static int fastopen_set_state(int fd, int state)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_SET_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
		.state = state
	};
	return fastopen_cmd(&cmd, NULL);
}

static int fastopen_cmp_set_state(int fd, int oldstate, int newstate)
{
	struct FASTOPENCmd cmd = {
		.op = FASTOPEN_CMP_SET_SOCK_STATE,
		.gsockfd = fastopen_to_global(fd),
		.cmp_state = oldstate,
		.state = newstate
	};
	struct FASTOPENRet ret;

	if (fastopen_cmd(&cmd, &ret) < 0 || !(ret.is_valid))
		return -1;
	return ret.state;
}

static void fastopen_change_state(int s, int new_state)
{
	ASSERT(s < MAX_SOCK_FD && s >= 0);
	fastopen_set_state(s, new_state);
}

static int fastopen_should_connect(int s)
{
	int old_state =
	    fastopen_cmp_set_state(s, FASTOPEN_SOCK_IDLE, FASTOPEN_SOCK_CONNECTING);
	return old_state == FASTOPEN_SOCK_IDLE;
}

static int fastopen_wait_for_connect(int s)
{
	int ret = 0;
	int cur_state = fastopen_wait_connect(s);
	if (cur_state == FASTOPEN_SOCK_INVALID)
		ret = -1;
	return ret;
}

static int fastopen_connect(int s, u8 * conn_data, int conn_data_len)
{
	struct fastopen_conn_data_req req = {
		.buf = conn_data,
		.len = conn_data_len
	};

	if (conn_data_len && real_setsockopt(s, IPPROTO_TCP,
					     TCP_FASTOPEN_CONN_DATA, &req,
					     sizeof(req)) < 0) {
		DEBUG("error in setting connection data\n");
		return -1;
	}

	struct sockaddr addr;
	socklen_t addrlen = sizeof(addr);

	//
	//    @FIXME:
	//            for ipv6 or unix domain socket, the size of sockaddr is not
	//            long enough :(
	//
	if (real_getsockopt(s, IPPROTO_TCP, TCP_FASTOPEN_CONN_ADDR, &addr, &addrlen)
	    < 0) {
		DEBUG("error in getting connection address\n");
		return -1;
	}

	if (real_connect(s, &addr, addrlen) < 0) {
		//
		//  for this kind of non-blocking situation, we have to wait for
		//  connection set up ourselves
		//
		if (errno == EINPROGRESS) {
			struct pollfd wait_set = {
				.fd = s,
				.events = POLL_OUT
			};
			int ret = poll(&wait_set, 1, 0);
			if (ret < 0
			    || (wait_set.revents & (POLLERR | POLLNVAL))) {
				goto err_conn;
			}
		} else {
err_conn:
			DEBUG("error in connecting: %s\n", strerror(errno));
			return -1;
		}
	}
	//
	//  Clean the special byte
	//
	fastopen_sock_table[s].no_special = 1;
	fastopen_change_state(s, FASTOPEN_SOCK_CONNECTED);
	return 0;
}

static int fastopen_get_special(int s, u8 * buf, int len)
{
	struct fastopen_conn_data_req req = {
		.buf = buf,
		.len = len
	};
	socklen_t reqlen = sizeof(req);
	if (real_getsockopt(s, IPPROTO_TCP, TCP_FASTOPEN_CONN_DATA, &req, &reqlen) <
	    0)
		return -1;
	//
	//  Whether the special data has been done
	//
	fastopen_sock_table[s].no_special = req.len <= len;
	return req.len;
}

static ssize_t fastopen_send(int s, const void *buf, size_t sz, int flags)
{
	int conn_size =
	    (int)(sz > FASTOPEN_MAX_CONN_DATA ? FASTOPEN_MAX_CONN_DATA : sz);
	u8 *conn_data = (u8 *) buf;
	if (fastopen_connect(s, conn_data, conn_size) < 0) {
		return -1;
	}

	int ret = conn_size;
	if (sz > conn_size) {
		int tmp = real_send(s, conn_data + conn_size, conn_size, flags);
		if (tmp < 0) {
			if (errno != EAGAIN && errno != EINTR)
				ret = -1;
		} else
			ret += tmp;
	}
	return ret;
}

static ssize_t fastopen_recv(int s, void *buf, size_t sz, int flags)
{
	size_t max_prefetch_len =
	    sz > FASTOPEN_MAX_CONN_DATA ? FASTOPEN_MAX_CONN_DATA : sz;
	ssize_t ret = fastopen_get_special(s, buf, max_prefetch_len);
	DEBUG("want %d, get %d\n", sz, ret);
	if (ret < (ssize_t) sz) {
		ssize_t tmp = real_recv(s, buf + ret, sz - ret, flags);
		DEBUG("get %d\n", tmp);
		if (tmp < 0) {
			if (errno != EAGAIN && errno != EINTR)
				ret = -1;
		} else
			ret += tmp;
	}
	return ret;
}

/*
 *  Auxiliary operations on complex msghdr's head
 *    @hdr        the operting msghdr
 *    @oplen      the requested pull or push length to the msghdr
 *    @(vect_pos, byte_pos) means the end position of the operation
 */
struct msghdr_op {
	const struct msghdr *hdr;
	int oplen;
	int vect_pos;
	int byte_pos;
};

#define MSGHDR_PULL_HEAD  1
#define MSGHDR_PUSH_HEAD  2

static void msghdr_getpos(struct msghdr_op *op)
{
	int vect, sum = 0;
	const struct msghdr *hdr = op->hdr;
	for (vect = 0, op->byte_pos = 0; vect < hdr->msg_iovlen; vect++) {
		if (hdr->msg_iov[vect].iov_len + sum > op->oplen) {
			op->byte_pos = op->oplen - sum;
			break;
		}
	}
	op->oplen = sum;
	op->vect_pos = vect;

}

static void msghdr_do_op(struct msghdr_op *op, u8 * buf, int opkind)
{
	int vect, sum;
	const struct msghdr *hdr = op->hdr;
	switch (opkind) {
	case MSGHDR_PUSH_HEAD:
		for (vect = 0, sum = 0; vect < op->vect_pos; vect++) {
			memcpy(hdr->msg_iov[vect].iov_base, buf + sum,
			       hdr->msg_iov[vect].iov_len);
			sum += hdr->msg_iov[vect].iov_len;
		}
		memcpy(hdr->msg_iov[vect].iov_base, buf + sum, op->byte_pos);
		break;
	case MSGHDR_PULL_HEAD:
		for (vect = 0, sum = 0; vect < op->vect_pos; vect++) {
			memcpy(buf + sum, hdr->msg_iov[vect].iov_base,
			       hdr->msg_iov[vect].iov_len);
			sum += hdr->msg_iov[vect].iov_len;
		}
		memcpy(buf + sum, hdr->msg_iov[vect].iov_base, op->byte_pos);
		break;
	}
}

/*
 * initialize a copy of the msghdr pointing to the data
 * that is left after operation. returns the total size
 * of splited message's body
 *
 * malloc called!
 */
static int msghdr_op_split(struct msghdr_op *op, struct msghdr *nmsg)
{
	int i, j;
	const struct msghdr *omsg = op->hdr;

	nmsg->msg_iovlen = 0;
	if (omsg->msg_iovlen == op->vect_pos) {
		return 0;
	}

	nmsg->msg_iovlen = omsg->msg_iovlen - op->vect_pos;
	nmsg->msg_name = omsg->msg_name;
	nmsg->msg_namelen = omsg->msg_namelen;
	nmsg->msg_control = omsg->msg_control;
	nmsg->msg_controllen = omsg->msg_controllen;
	nmsg->msg_flags = omsg->msg_flags;
	nmsg->msg_iov = malloc(sizeof(struct iovec) * nmsg->msg_iovlen);
	if (!nmsg->msg_iov) {
		return -1;
	}
	//
	//  Deal with special message header
	//
	nmsg->msg_iov[0].iov_len =
	    omsg->msg_iov[op->vect_pos].iov_len - op->byte_pos;
	nmsg->msg_iov[0].iov_base =
	    omsg->msg_iov[op->vect_pos].iov_base + op->byte_pos;

	int sum = nmsg->msg_iov[0].iov_len;
	for (i = 1, j = op->vect_pos + 1; j < omsg->msg_iovlen; i++, j++) {
		nmsg->msg_iov[i] = omsg->msg_iov[j];
		sum += omsg->msg_iov[j].iov_len;
	}
	return sum;
}

/*
 *  We only send out the first segment in sendmsg! This is to simplify the
 *  programming.
 */
static ssize_t fastopen_sendmsg(int s, const struct msghdr *msg, int flags)
{
	ASSERT(msg && msg->msg_iovlen > 0);

	u8 special_buf[FASTOPEN_MAX_CONN_DATA];
	struct msghdr nhdr;
	struct msghdr_op op = {
		.hdr = msg,
		.oplen = FASTOPEN_MAX_CONN_DATA,
	};

	msghdr_getpos(&op);
	msghdr_do_op(&op, special_buf, MSGHDR_PULL_HEAD);
	if (fastopen_connect(s, special_buf, op.oplen) < 0)
		return -1;

	ssize_t ret = op.oplen;
	if (msghdr_op_split(&op, &nhdr)) {
		ssize_t tmp = real_sendmsg(s, &nhdr, flags);
		if (tmp < 0) {
			if (errno != EAGAIN && errno != EINTR)
				ret = -1;
		} else {
			ret += tmp;
		}
		free(nhdr.msg_iov);
	}
	return ret;
}

static ssize_t fastopen_recvmsg(int s, const struct msghdr *msg, int flags)
{
	struct msghdr_op op = {
		.oplen = FASTOPEN_MAX_CONN_DATA,
		.hdr = msg
	};

	u8 special_buf[FASTOPEN_MAX_CONN_DATA];
	msghdr_getpos(&op);
	ssize_t ret = fastopen_get_special(s, special_buf, op.oplen);
	msghdr_do_op(&op, special_buf, MSGHDR_PUSH_HEAD);

	if (ret >= op.oplen) {
		return ret;
	}

	struct msghdr nhdr;
	if (msghdr_op_split(&op, &nhdr)) {
		ssize_t tmp = real_recvmsg(s, &nhdr, flags);
		if (tmp < 0) {
			if (errno != EAGAIN && errno != EINTR)
				ret = -1;
		} else
			ret += tmp;
		free(nhdr.msg_iov);
	}
	return ret;
}

/***************************************************************
 *                Slow Connection Management                   *
 *-------------------------------------------------------------*
 *      deal with connect method called without sending        *
 *      something                                              *
 ***************************************************************/
/*
 *  Slow socket list and locks
 */
struct slow_sock {
	int sock_fd;
	int timeo_tick;		/* how many before timeout */
	struct slow_sock *next;
} slow_sock_head, *slow_sock_list;

/*
 *  Global spin lock to protect the slow sock list
 */
static pthread_spinlock_t slow_sock_lock;

/*
 *  We want the slow sock list to be protected against fork behavior
 */
static pthread_mutex_t fork_protect_lock = PTHREAD_MUTEX_INITIALIZER;

/*
 *  Timer ID and tick configure
 */
static timer_t slow_conn_timer;
static double slow_conn_tick_time = 100000;	/* 100 us per tick */
static double slow_conn_timeout_tick = 2;	/* time out after 1-2 ticks */

/*
 *  Force sending out the connect frame. This might cause some performance
 *  panelty in way that server first send out its data and client waits
 *  for that!
 */
static void slow_connect(union sigval unused_val)
{
	pthread_mutex_lock(&fork_protect_lock);

	struct slow_sock *list = slow_sock_list;
	pthread_spin_lock(&slow_sock_lock);
	//printf("active slow connect\n");
	while (list->next) {
		//
		//  @Lock hold:
		//    . spin slow_sock_lock
		//
		struct slow_sock *sock = list->next;
		if (--sock->timeo_tick > 0) {
			list = list->next;
			continue;
		}
		DEBUG("in slow conn\n");
		list->next = sock->next;
		pthread_spin_unlock(&slow_sock_lock);

		//
		//  @Critical section:
		//    @potential competitors:
		//      . the one that calls on sending routings
		//    @shared vars:
		//      . fastopen_sock_table entry's state 
		//
		//  fastopen_should_connect use atomic operation to ensure
		//  there is only one that does the connect work. For
		//  the slow connect thread, there is no need to wait
		//  for other ones to finish the connect, and it can
		//  go on to next socket fd.
		//
		if (fastopen_should_connect(sock->sock_fd)) {
			//
			//        The socket state will be changed to connected in fastopen_connect.
			//        Or if some error occurs, change the state to invalid
			//
			if (fastopen_connect(sock->sock_fd, NULL, 0) < 0) {
				DEBUG("error in do connection: %s\n",
				      strerror(errno));
				errno = ECONNRESET;
				fastopen_change_state(sock->sock_fd,
						  FASTOPEN_SOCK_INVALID);
			}
		}
		//
		//  Release the resources and go on to next sock
		//
		free(sock);
		pthread_spin_lock(&slow_sock_lock);
	}
	pthread_spin_unlock(&slow_sock_lock);
	pthread_mutex_unlock(&fork_protect_lock);
}

/*
 *  Slow connection startup and finalize configure
 */

static void setup_slow_timer()
{
	struct sigevent ev = {
		.sigev_notify = SIGEV_THREAD,
		.sigev_notify_attributes = NULL,
		.sigev_notify_function = &slow_connect,
	};

	if (timer_create(CLOCK_REALTIME, &ev, &slow_conn_timer) < 0) {
		fprintf(stderr, "can not handle connection timer\n");
		abort();
	}
	struct itimerspec tick_spec = {
		.it_value = {
			     .tv_sec = 0,
			     .tv_nsec = slow_conn_tick_time * 1000},
		.it_interval = {
				.tv_sec = 0,
				.tv_nsec = slow_conn_tick_time * 1000},
	};
	if (timer_settime(slow_conn_timer, 0, &tick_spec, NULL) < 0) {
		fprintf(stderr, "can not handle connection timer\n");
		abort();
	}
}

static void setup_slow_list()
{
	slow_sock_list = &slow_sock_head;
	slow_sock_list->next = NULL;
	pthread_spin_init(&slow_sock_lock, PTHREAD_PROCESS_PRIVATE);
}

static void setup_sock_table()
{
	int i;
	for (i = 0; i < MAX_SOCK_FD; i++) {
		fastopen_sock_table[i].type = FASTOPEN_NONE;
		fastopen_sock_table[i].global_id = FASTOPEN_NEW_SOCK;
		fastopen_sock_table[i].no_special = 0;
	}
}

static void slow_conn_init()
{
	setup_slow_timer();
	setup_slow_list();
	setup_sock_table();
}

static void slow_conn_destroy()
{
	timer_delete(slow_conn_timer);
	//
	//  Although it's quite brute to hold spin lock for long,
	//  we have come to the end of the program, and just let
	//  it go :)
	//
	pthread_spin_lock(&slow_sock_lock);
	while (slow_sock_list->next) {
		struct slow_sock *tmp = slow_sock_list->next;
		slow_sock_list->next = tmp->next;
		free(tmp);
	}
	pthread_spin_unlock(&slow_sock_lock);
}

static int slow_sock_list_add(int s)
{
	struct slow_sock *sock = malloc(sizeof(struct slow_sock));
	if (!sock)
		return -1;
	sock->sock_fd = s;
	sock->timeo_tick = slow_conn_timeout_tick;

	pthread_mutex_lock(&fork_protect_lock);
	pthread_spin_lock(&slow_sock_lock);
	sock->next = slow_sock_list->next;
	slow_sock_list->next = sock;
	pthread_spin_unlock(&slow_sock_lock);
	pthread_mutex_unlock(&fork_protect_lock);
	return 0;
}

/*
 *	Fork protector
 */
static void parent_before_fork()
{
	DEBUG("parent before fork\n");
	pthread_mutex_lock(&fork_protect_lock);
}

static void parent_after_fork()
{
	pthread_mutex_unlock(&fork_protect_lock);
	DEBUG("parent after fork\n");
}

static void child_on_fork()
{
	//
	//      Workaround to solve the lock state:
	//
	//      1. use of spinlock is in timer thread and connect wrapper
	//      2. later can not be fork point
	//      3. former is protected by fork protect lock
	//
	pthread_mutex_unlock(&fork_protect_lock);
	//pthread_spin_destroy(&slow_sock_lock);
	//pthread_spin_init(&slow_sock_lock, PTHREAD_PROCESS_PRIVATE);

	//
	//      Send command add proc
	//      Send command dup proc
	//
	fastopen_add_proc();
	fastopen_dup_proc();
	setup_slow_timer();
	DEBUG("child after fork\n");
}


/***************************************************************
 *                Library startup			       *
 *-------------------------------------------------------------*
 *	All the library init functions go here.		       *
 ***************************************************************/

/*
 *	Work around to some ugly cases that this mechanism does 
 *	not work well
 */
static int has_inited = 0;

static __attribute__ ((constructor))
void fastopen_wrap_constructor()
{
	//libc_init();
	slow_conn_init();
	fastopen_add_proc();
	pthread_atfork(parent_before_fork, parent_after_fork, child_on_fork);
	has_inited = 1;
	//logs();
}

static inline void ensure_init()
{
#if 0
	if (unlikely(!has_inited)) {
		fastopen_wrap_constructor();
		has_inited = 1;
	}
#endif
}

/***************************************************************
 *                Redirected Syscalls                          *
 *-------------------------------------------------------------*
 *      syscalls are redirected, mainly for socket, connect,   *
 *      send and recv. No AIO related implementation now       *
 ***************************************************************/
static int is_stream(int s)
{
	int so_type = -1;
	socklen_t typelen = sizeof(so_type);
	if (real_getsockopt(s, SOL_SOCKET, SO_TYPE, &so_type, &typelen) < 0) {
		DEBUG("can not get socket tpye!\n");
		return 0;
	}
	return so_type == SOCK_STREAM;
}

static int is_socket(int fd)
{
	struct stat fd_stat;
	if (fstat(fd, &fd_stat) < 0) {
		DEBUG("can not stat fd: %d, %s\n", fd, strerror(errno));
		return 0;
	}
	return S_ISSOCK(fd_stat.st_mode);
}

static int is_fastopen_sock(int fd)
{
	return fastopen_sock_table[fd].type & FASTOPEN_SOCK;
}

static int is_fastopen_client(int fd)
{
	return fastopen_sock_table[fd].type & FASTOPEN_CLIENT;
}

static int is_fastopen_server(int fd)
{
	return fastopen_sock_table[fd].type & FASTOPEN_SERVER;
}

int socket(int family, int type, int protocol)
{
	ensure_init();
	int sfd = real_socket(family, type, protocol);
	int fastopen_on = 1;
	fastopen_sock_table[sfd].type = 0;

	if (family != AF_INET || (type & SOCK_STREAM) == 0)
		return sfd;
	DEBUG("create a socket: %d %d %d\n", family, type, sfd);
	if (real_setsockopt(sfd, IPPROTO_TCP, TCP_FASTOPEN, &fastopen_on, sizeof(int)) <
	    0)
		return -1;
	//
	//  Ensure this socket fd is not allocated yet!
	//
	fastopen_sock_table[sfd].type = FASTOPEN_SOCK;
	fastopen_add_sock(sfd);
	DEBUG("finish a socket\n");
	return sfd;
}

int connect(int s, const struct sockaddr *addr, socklen_t len)
{
	ensure_init();
	if (!is_fastopen_sock(s))
		return real_connect(s, addr, len);
	printf("connect addr %x, %u\n", 
			((struct sockaddr_in*)addr)->sin_addr.s_addr,
			ntohs(((struct sockaddr_in*)addr)->sin_addr.s_addr));

	if (real_setsockopt(s, IPPROTO_TCP, TCP_FASTOPEN_CONN_ADDR, addr, len))
		return -1;
	if (slow_sock_list_add(s) < 0)
		return -1;
	// Don't change state here, leave it later
	fastopen_sock_table[s].type |= FASTOPEN_CLIENT;
	fastopen_sock_table[s].no_special = 0;
	return 0;
}

ssize_t send(int s, const void *buf, size_t len, int flags)
{
	ensure_init();
	if (!is_fastopen_client(s)) {
		return real_send(s, buf, len, flags);
	}

	if (likely(fastopen_sock_table[s].no_special)) {
		return real_send(s, buf, len, flags);
	} else {
		//
		//  @Critical section:
		//    @potential competitors:
		//      # the slow connect daemon thread
		//    @shared vars:
		//      # fastopen_sock_table entry's state
		//
		//  The sender is a little different from daemon timer thread,
		//  since it must complete the actual sending. So it must wait
		//  for connection complete to call send, rather than to return
		//  directly.
		//
		if (fastopen_should_connect(s)) {
			return fastopen_send(s, buf, len, flags);
		} else {
			if (fastopen_wait_for_connect(s) < 0)
				return -1;
			return real_send(s, buf, len, flags);
		}
	}
}

ssize_t sendmsg(int s, const struct msghdr * msg, int flags)
{
	ensure_init();
	if (!is_fastopen_client(s)) {
		return real_sendmsg(s, msg, flags);
	}

	if (likely(fastopen_sock_table[s].no_special)) {
		return real_sendmsg(s, msg, flags);
	} else {
		if (fastopen_should_connect(s)) {
			return fastopen_sendmsg(s, msg, flags);
		} else {
			if (fastopen_wait_for_connect(s) < 0)
				return -1;
			return real_sendmsg(s, msg, flags);
		}
	}
}

ssize_t write(int s, const void *buf, size_t len)
{
	ensure_init();
	if (!is_fastopen_sock(s))
		return real_write(s, buf, len);
	return send(s, buf, len, 0);
}

ssize_t writev(int s, const struct iovec * vec, int veclen)
{
	ensure_init();
	if (!is_fastopen_sock(s))
		return real_writev(s, vec, veclen);
	struct msghdr hdr = {
		.msg_name = NULL,
		.msg_namelen = 0,
		.msg_iov = (struct iovec *)vec,
		.msg_iovlen = veclen,
		.msg_control = NULL,
		.msg_controllen = 0,
		.msg_flags = 0
	};
	return sendmsg(s, &hdr, 0);
}

ssize_t recv(int s, void *buf, size_t len, int flags)
{
	ensure_init();
	if (!is_fastopen_server(s))
		return real_recv(s, buf, len, flags);

	if (likely(fastopen_sock_table[s].no_special)) {
		return real_recv(s, buf, len, flags);
	} else {
		return fastopen_recv(s, buf, len, flags);
	}
}

ssize_t read(int s, void *buf, size_t len)
{
	ensure_init();
	if (!is_fastopen_sock(s))
		return real_read(s, buf, len);
	return recv(s, buf, len, 0);
}

ssize_t recvmsg(int s, struct msghdr * msg, int flags)
{
	ensure_init();
	if (!is_fastopen_server(s))
		return real_recvmsg(s, msg, flags);

	if (likely(fastopen_sock_table[s].no_special)) {
		return real_recvmsg(s, msg, flags);
	} else {
		return fastopen_recvmsg(s, msg, flags);
	}
}

ssize_t readv(int s, const struct iovec * vec, int veclen)
{
	ensure_init();
	if (!is_fastopen_sock(s))
		return real_readv(s, vec, veclen);
	struct msghdr hdr = {
		.msg_name = NULL,
		.msg_namelen = 0,
		.msg_iov = (struct iovec *)vec,
		.msg_iovlen = veclen,
		.msg_control = NULL,
		.msg_controllen = 0,
		.msg_flags = 0
	};
	return recvmsg(s, &hdr, 0);
}

int accept(int s, struct sockaddr *addr, socklen_t * addrlen)
{
	ensure_init();
	DEBUG("before accept\n");
	int serv_fd = real_accept(s, addr, addrlen);
	DEBUG("after accept, %d\n", serv_fd);
	if (serv_fd < 0) {
		return -1;
	}
	//
	//    notify this sock to FASTOPEN daemon? Maybe not, since it
	//    does not have connect state
	//
	fastopen_sock_table[serv_fd].type = FASTOPEN_SOCK | FASTOPEN_SERVER;
	fastopen_sock_table[serv_fd].no_special = 0;
	return serv_fd;
}

int accept4(int s, struct sockaddr *addr, socklen_t * addrlen, int flag)
{
	DEBUG("before accept\n");
	ensure_init();
	int serv_fd = real_accept4(s, addr, addrlen, flag);
	DEBUG("after accept, %d %s\n", serv_fd, strerror(errno));
	if (serv_fd < 0) {
		return -1;
	}
	//
	//    notify this sock to FASTOPEN daemon? Maybe not, since it
	//    does not have connect state
	//

	fastopen_sock_table[serv_fd].type = FASTOPEN_SOCK | FASTOPEN_SERVER;
	fastopen_sock_table[serv_fd].no_special = 0;
	return serv_fd;
}

int close(int s)
{
	ensure_init();
	if (!is_fastopen_sock(s)) {
		if (s != 2) 
			return real_close(s);
		else 
			return 0;
	}
	//
	//    FIXME:
	//    Possible race is that the socket is set to invalid and freed, but it's
	//    later allocated again and returns an invalid state.
	//
	fastopen_sock_table[s].type = 0;
	fastopen_sock_table[s].no_special = 0;
	fastopen_change_state(s, FASTOPEN_SOCK_INVALID);
	fastopen_del_sock(s);
	return real_close(s);
}

int getpeername(int s, struct sockaddr *addr, socklen_t * addrlen)
{
	ensure_init();
	if (!is_fastopen_sock(s))
		return real_getpeername(s, addr, addrlen);

	struct sockaddr conn_addr;
	socklen_t conn_addrlen = sizeof(conn_addr);

	//
	//    @FIXME:
	//            see note in fastopen_connect
	//
	if (real_getsockopt
	    (s, IPPROTO_TCP, TCP_FASTOPEN_CONN_ADDR, &conn_addr,
	     &conn_addrlen) < 0) {
		DEBUG("error in getting connection address\n");
		return -1;
	}
	if (*addrlen < conn_addrlen)
		*addrlen = conn_addrlen;
	memcpy(addr, &conn_addr, *addrlen);
	return 0;
}

/*
int open(const char* path, int flags, mode_t m)
{
	ensure_init();
	int fd = real_open(path, flags, m);
	fastopen_sock_table[fd].type = 0;
	return fd;
}

int creat(const char* path, mode_t m)
{
	int fd = real_creat(path, m);
	fastopen_sock_table[fd].type = 0;
	return fd;
}

*/
