
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <poll.h>
#include <sys/select.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <fcntl.h>

static int inline stable_send(int fd, const char* buf, int len)
{
  int head = 0;
  while (len) {
    int tmp = send(fd, buf + head, len, MSG_NOSIGNAL);
    if (tmp < 0) {
      if (errno == EINTR || errno == EAGAIN) continue;
      else if (errno == EPIPE) {
        printf("other side closed: broken pipe\n");
        return 0;
      } else {
        printf("error! %s\n", strerror(errno));
        return -1;
      }
    }
    len -= tmp;
    head += tmp;
  }
  return head - len;
}

static int inline stable_recv(int fd, char* buf, int len)
{
  int head = 0;
  while (len) {
    int tmp = recv(fd, buf + head, len, 0);
    if (tmp < 0) {
      if (errno == EINTR || errno == EAGAIN) {
        continue;
      } else {
        printf("error! %s\n", strerror(errno));
        return -1;
      }
    } else if (tmp == 0) {
      printf("other side has closed\n");
      return 0;
    }
    len -= tmp;
    head += tmp;
  }
  return head - len;
}

static int inline create_fastopen_server()
{
  int serv = socket(AF_INET, SOCK_STREAM, 0);
  int fastopen_type = 1;
  int reuse = 1;
  assert(serv >= 0);
  assert(setsockopt(serv, IPPROTO_TCP, TCP_FASTOPEN, &fastopen_type, sizeof(int)) == 0);
  assert(setsockopt(serv, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) == 0);
  struct sockaddr_in addr = {
    .sin_family = AF_INET,
    .sin_port = htons(10000),
    .sin_addr = {
      .s_addr = htonl(INADDR_ANY)
    }
  };
  assert(bind(serv, (struct sockaddr*) &addr, sizeof(addr)) == 0);
  assert(listen(serv, 100) == 0);
  fprintf(stderr, "fastopen server created\n");
  return serv;
}


static int inline create_fastopen_client(char* servaddr, char* pref_msg)
{
  int fastopen_type = 1;
  int c = socket(AF_INET, SOCK_STREAM, 0);
  assert(c >= 0);
  assert(setsockopt(c, IPPROTO_TCP, TCP_FASTOPEN, &fastopen_type, sizeof(int)) == 0);
  struct fastopen_conn_data_req req = {
    .buf = (unsigned char*) pref_msg,
    .len = strlen(pref_msg) + 1 // add the end of the string
  };
//  printf("%d\n", req.len);
  assert(setsockopt(c, IPPROTO_TCP, TCP_FASTOPEN_CONN_DATA, &req, sizeof(req)) == 0);

  struct sockaddr_in addr = {
    .sin_family = AF_INET,
    .sin_port = htons(10000),
    .sin_addr = {
      .s_addr = inet_addr(servaddr)
    }
  };
  
  if (connect(c, (struct sockaddr*) &addr, sizeof(addr)) < 0) {
	  printf("fail in connecting: %d: %s\n", c, strerror(errno));
  }
  fprintf(stderr, "fastopen client created\n");
  return c;
}

static void server_loop(int serv, int (*fn)(int fd))
{
  while (1) {
    struct sockaddr_in caddr;
    socklen_t caddrlen = sizeof(caddr);
    int c = accept(serv, (struct sockaddr*) &caddr, &caddrlen);
    assert(c > 0);
    if (fn(c) != 0) {
      printf("error in functioning\n");
      break;
    }
  }
  close(serv);
}


static int test_select(int fd)
{
	fd_set test_set;
	FD_ZERO(&test_set);
	FD_SET(fd, &test_set);
	struct timeval timeo = {
		.tv_sec = 1,
		.tv_usec = 0
	};
	int val = select(fd + 1, &test_set, NULL, NULL, &timeo);
	if (val < 0) {
		/*
		 *	Timeout
		 */
		printf("error in select: %s\n", strerror(errno));
		abort();
	}
	return val;
}

static int test_poll(int fd)
{

	struct pollfd pfd = {
		.fd = fd,
		.events = POLLIN
	};

	int val = poll(&pfd, 1, 1000);
	if (val < 0) {
		if (errno == EAGAIN)
			return -1;
		printf("error in poll: %s\n", strerror(errno));
		abort();
	}
	return val;
}

static int test_epoll(int fd)
{
	int ep = epoll_create(1);
	struct epoll_event ev = {
		.events = EPOLLIN,
		.data = {
			.fd = fd
		}
	};
	if (epoll_ctl(ep, EPOLL_CTL_ADD, fd, &ev) < 0){
		fprintf(stderr, "can not create epoll fd\n");
		abort();
	}
	int val = epoll_wait(ep, &ev, 1, 1000);
	if (val < 0) {
		if (errno == EAGAIN)
			return -1;
		printf("error in epoll: %s\n", strerror(errno));
	}
	close(ep);
	return val;
}

static inline int create_std_client(char* ip)
{
	int s = socket(AF_INET, SOCK_STREAM, 0);
	struct sockaddr_in addr = {
		.sin_family = AF_INET,
		.sin_port = htons(10000),
		.sin_addr = {
			.s_addr = inet_addr(ip)
		}
	};
	return connect(s, (struct sockaddr*) &addr, sizeof(addr)) == 0?  s: -1;
}

static inline int create_std_server()
{
	int s = socket(AF_INET, SOCK_STREAM, 0);
	struct sockaddr_in addr = {
		.sin_family = AF_INET,
		.sin_port = htons(10000),
		.sin_addr = {
			.s_addr = INADDR_ANY
		}
	};
	int reuse = 1;
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) < 0) {
		printf("error in set reuse address\n");
		return -1;
	}
	if (bind(s, (struct sockaddr*) &addr, sizeof(addr)) < 0) {
		printf("can not bind\n");
		return -1;
	}
	return listen(s, 100) == 0? s: -1;
}


static inline int stable_send1(int fd, char* buf, int size)
{

}

static inline int stable_recv1(int fd, char* buf, int size)
{

}


static int inline create_fastopen_server_nb()
{
  int serv = socket(AF_INET, SOCK_STREAM, 0);
  int fastopen_type = 1;
  int reuse = 1;
  assert(serv >= 0);
  assert(setsockopt(serv, IPPROTO_TCP, TCP_FASTOPEN, &fastopen_type, sizeof(int)) == 0);
  assert(setsockopt(serv, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) == 0);
  assert(fcntl(serv, F_SETFD, O_NONBLOCK) == 0);
  struct sockaddr_in addr = {
    .sin_family = AF_INET,
    .sin_port = htons(10000),
    .sin_addr = {
      .s_addr = htonl(INADDR_ANY)
    }
  };
  assert(bind(serv, (struct sockaddr*) &addr, sizeof(addr)) == 0);
  assert(listen(serv, 100) == 0);
  fprintf(stderr, "fastopen server created\n");
  return serv;
}


static int inline create_fastopen_client_nb(char* servaddr, char* pref_msg)
{
  int fastopen_type = 1;
  int c = socket(AF_INET, SOCK_STREAM, 0);
  assert(c >= 0);
  assert(setsockopt(c, IPPROTO_TCP, TCP_FASTOPEN, &fastopen_type, sizeof(int)) == 0);
  struct fastopen_conn_data_req req = {
    .buf = (unsigned char*) pref_msg,
    .len = strlen(pref_msg) + 1 // add the end of the string
  };
  assert(fcntl(c, F_SETFD, O_NONBLOCK) == 0);
//  printf("%d\n", req.len);
  assert(setsockopt(c, IPPROTO_TCP, TCP_FASTOPEN_CONN_DATA, &req, sizeof(req)) == 0);

  struct sockaddr_in addr = {
    .sin_family = AF_INET,
    .sin_port = htons(10000),
    .sin_addr = {
      .s_addr = inet_addr(servaddr)
    }
  };
  if (connect(c, (struct sockaddr*) &addr, sizeof(addr)) < 0) {
	  printf("in connecting: %d: %s\n", c, strerror(errno));
  }
  fprintf(stderr, "fastopen client created\n");
  return c;
}
